package CryptoPhotoUtils;

use warnings;
use strict;
use Carp;
use LWP::UserAgent;
use Digest::HMAC_SHA1 qw(hmac_sha1_hex);
use JSON;
use Digest::SHA qw(sha256_base64);

our $VERSION = '1.20160609';

# CryptoPhoto API Endpoint URL
use constant CRYPTO_SERVER => 'cryptophoto.com/api/';
use constant SERVER_ERROR  => 'error-cryptophoto-not-reachable';

sub new {
  my($class, $server, $privatekey, $publickey, $uid) = @_;       

  croak "private key missing" unless $privatekey;
  croak "public key missing" unless $publickey;
  croak "user id missing" unless $uid;

  my $self = {
    server     => $server ? $server . "/api/" : "https://" . CRYPTO_SERVER,
    privatekey => $privatekey,
    publickey  => $publickey,
    uid        => $uid
  };   
  bless($self, $class);

  return $self;
}


#HTTP[S] CryptoPhoto API call
sub post_request {
  my $self = shift;
  my ($url, $args) = @_;

  my $ua = LWP::UserAgent->new();
  return $ua->post($url, $args);
}


#request a new CryptoPhoto session
sub start_session {
  my($self, $ip, $authentication) = @_;

  my $time = time;
  my $sig = $self->_make_signature($time);

  my $response = $self->post_request(
        $self->{server} . "get/session",
        {
          "publickey" => $self->{"publickey"},
          "uid"       => $self->{"uid"},
          "time"      => $time,
          "signature" => $sig,
          "ip"        => $ip,
          "authentication" => $authentication ? "true" : "false"
        });  



  #check response
  if ($response->is_success) {
    my ($answer, $message, $hastoken) = split(/\n/, $response->content);
    if ($answer =~ /success/ ) {
      $self->{'sid'} = $message;
      return { is_valid => 1, sid => $message, hastoken => defined $hastoken ? 1 : 0 };
    }
    else {
      chomp $message;
      return { is_valid => 0, error => $message };
    }
  }
  else {#server unreachable or bad request
      return { is_valid => 0, error => SERVER_ERROR };
  }
}


#generate javascript code for getting/managing tokens
sub get_gen_widget {
  my $self = shift;

  if (exists $self->{'sid'}) {
    return '<script type="text/javascript" src="'.$self->{server}.'token?sd='.$self->{'sid'}.'"></script>';
  }
  else {
    croak "Generate session first";
  }
}


#generate javascript code for getting a challenge  
sub get_auth_widget {
  my $self = shift;

  if (exists $self->{'sid'}) {
    return '<script type="text/javascript" src="'.$self->{server}.'challenge?sd='.$self->{'sid'}.'"></script>';
  }
  else {
    croak "Generate session first";
  }
}


#post and check a user's response for a given challenge
sub verify_response {

  my($self, $selector, $response_row, $response_col, $cp_phc, $ip) = @_;
  
  croak "selector missing" if(!$self->_defined($selector));
  croak "response missing" if((!$self->_defined($response_col) || !$self->_defined($response_row)) && !$self->_defined($cp_phc));

  my $time = time;
  my $sig = $self->_make_signature($time);
  
  #post user data and his challenge response
  my $resp = $self->post_request(
        $self->{server}."verify",
        {
          "publickey"    => $self->{"publickey"},
          "uid"          => $self->{"uid"},
          "response_row" => $response_row, 
          "response_col" => $response_col,
          "selector"     => $selector,
          "cph"          => $cp_phc, 
          "ip"           => $ip,
          "time"         => $time,
          "signature"    => $sig
        });

  if ($resp->is_success) {
    my ($answer, $message, $description) = split(/\n/, $resp->content);
    if ($answer =~ /success/ ) {
      #challenge matching successful
      return { is_valid => 1, message => $message };
    }
    else {
      #challenge matching error
      return { is_valid => 0, error => $message, description => $description };
    }
  }
  else {#server unreachable or bad request
    return { is_valid => 0, error => SERVER_ERROR };
  }
}

sub verify_cptv_response {
  my ($parms) = @_;

  my $res = {};

  if (defined $parms->{'cpJWSrfc7515'}) {

    # let's check the cptv JWS token by querying CP
    my $ua = LWP::UserAgent->new();

    my $resp = $ua->post($self->{server}."verify/cptv.json",
                         { "token" => $parms->{'cpJWSrfc7515'}});

    if ($resp->is_success) {

      my $answer = fromJSON(undef, $resp->content);

      if (ref($answer) eq "HASH" && scalar(keys %$answer)) {

        if (defined $answer->{'success'}) {

          my $token = $parms->{'cpJWSrfc7515'};
          my @splits = split(/\./, $token, 3);

          my $claim = fromJSON(undef, decode_base64url($splits[1]));

          if (ref($claim) eq "HASH" && scalar(keys %$claim)) {

            my $fieldsOrder = $claim->{'fieldsOrder'};
            my $fieldsSha256 = $claim->{'fieldsSha256'};

            my @fields = split(/,/,$fieldsOrder); my $shacontent = "";
            foreach my $field (@fields) {
              $shacontent .= $parms->{$field} if(defined $parms->{$field});
            }

            if ($fieldsSha256 eq sha256_base64($shacontent)) {
              return { is_valid => 1 };
            } else {
              return { is_valid => 0, error => "POSTed field values have been changed" };
            }

          }
        } else {
          return { is_valid => 0, error => $answer->{'description'}};
        }
      } else {
         return { is_valid => 0, error => "CRYPTOPHOTO responded with invalid format" } ;
      }
    } else {
      return { is_valid => 0, error => SERVER_ERROR };
    }
  }

  return { is_valid => 0, error => "JWT token not provided" };
}


# internal: check if a variable is defined and has a value
sub _defined {
  my ($self, $val) = @_;

  if(defined $val && $val ne "") {
    return 1;
  }
  return 0;
}


# internal: create an api call signature
sub _make_signature {
  my($self, $time) = @_;
  return hmac_sha1_hex($self->{"privatekey"} . $time . $self->{"uid"} . $self->{"publickey"}, $self->{"privatekey"});
}

sub fromJSON {
  my ($self, $txt) = @_;
  my %json;
  eval('%json = %{decode_json($txt)}');
  if($@) {
    my $empty = {};
    return $empty;
  }
  return \%json;
}

sub toJSON {
  my ($self, $ref) = @_;
  return encode_json($ref);
}


1; 

=head1 NAME

CryptoPhotoUtils - Helper module for CryptoPhoto API

=head1 SYNOPSIS

 use CryptoPhotoUtils;
 
 my $server = "https://cryptophoto.com";
 my $cp = CryptoPhotoUtils->new($server, $privatekey, $publickey, $userid);

 my $response = $cp->start_session($ip);

 if ($response->{"is_valid"}) {
   my $seesion_id = $response->{"sid"};
   my $has_token = $response->{"hastoken"};
 }
 
 my $js_html = $cp->get_gen_widget();
 my $js_html = $cp->get_auth_widget();

 my $response = $cp->verify_response($selector, $response_row, $response_col, $cp_phc, $ip);
 my $response = $cp->send_tv($callback_url, $json, $ip);

=head1 DESCRIPTION

  This is a Perl library that handles calling L<CryptoPhoto API|https://cryptophoto.com/api>.
     - Main Page
         http://cryptophoto.com/
     - About Cryptophoto
         http://cryptophoto.com/about
     - Register to CryptoPhoto
         http://cryptophoto.com/demo/register/


=head1 CONSTRUCTOR METHODS

=over 4

=item $cp = CryptoPhotoUtils->new( $privatekey, $publickey, $userid )

This method constructs a new C<CryptoPhotoUtils> object and returns it.
L<CryptoPhoto API|https://cryptophoto.com/api> contains info on how to obtain API keys.

=back

=head1 API METHODS

The methods described in this section are used to dispatch API calls to L<CryptoPhoto|https://cryptophoto.com>
Service Endpoint. The following methods are provided:

=over

=item $cp->start_session( $ip )

Initiate a CP API session.

   $ip  IP of the user for whom the API call is made

Allways call this method before calling any other API methods.

=item $js_html = $cp->get_gen_widget()

See <CryptoPhoto API|https://cryptophoto.com/admin/api>

=item $js_html = $cp->get_auth_widget()

See <CryptoPhoto API|https://cryptophoto.com/admin/api>

=item $response = $cp->verify_response($selector, $response_row, $response_col, $cp_phc, $ip)

See <CryptoPhoto API|https://cryptophoto.com/admin/api>

=back

=head1 AUTHOR

L<CryptoPhoto|https://cryptophoto.com> E<lt>tech@cryptophoto.comE<gt>

=head1 COPYRIGHT

Copyright 2016 - L<CryptoPhoto|https://cryptophoto.com>

=head1 LICENSE

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

=cut
